#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2015 Hamilton Kibbe <ham@hamiltonkib.be>
# Copyright 2022 Jan Götte <code@jaseg.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
from ..cam import FileSettings


def test_zero_suppression():
    # Default format
    settings = FileSettings(number_format=(2,5), zeros='leading')
    test_cases = [
        ("1", 0.00001),
        ("10", 0.0001),
        ("100", 0.001),
        ("1000", 0.01),
        ("10000", 0.1),
        ("100000", 1.0),
        ("1000000", 10.0),
        ("0", 0.0),
    ]

    assert settings.write_gerber_value(0.000000001) == "0"

    for string, value in test_cases:
        assert value == settings.parse_gerber_value(string)
        assert string == settings.write_gerber_value(value)

    settings = FileSettings(number_format=(2,5), zeros='trailing')
    test_cases = [
        ("1", 10.0),
        ("01", 1.0),
        ("001", 0.1),
        ("0001", 0.01),
        ("00001", 0.001),
        ("000001", 0.0001),
        ("0000001", 0.00001),
        ("0", 0.0),
    ]

    assert settings.write_gerber_value(0.000000001) == "0"

    for string, value in test_cases:
        assert value == settings.parse_gerber_value(string)
        assert string == settings.write_gerber_value(value)


def test_format():
    test_cases = [
        ((2, 7), "1", 0.0000001),
        ((2, 6), "1", 0.000001),
        ((2, 5), "1", 0.00001),
        ((2, 4), "1", 0.0001),
        ((2, 3), "1", 0.001),
        ((2, 2), "1", 0.01),
        ((2, 1), "1", 0.1),
        ((2, 6), "0", 0),
    ]
    for fmt, string, value in test_cases:
        settings = FileSettings(number_format=fmt, zeros='leading')
        assert value == settings.parse_gerber_value(string)
        assert string == settings.write_gerber_value(value)

    test_cases = [
        ((6, 5), "1", 100000.0),
        ((5, 5), "1", 10000.0),
        ((4, 5), "1", 1000.0),
        ((3, 5), "1", 100.0),
        ((2, 5), "1", 10.0),
        ((1, 5), "1", 1.0),
        ((2, 5), "0", 0),
    ]
    for fmt, string, value in test_cases:
        settings = FileSettings(number_format=fmt, zeros='trailing')
        assert value == settings.parse_gerber_value(string)
        assert string == settings.write_gerber_value(value)

def test_parse_format_validation():
    for fmt in (7,5), (5,8), (13,1):
        with pytest.raises(ValueError):
            settings = FileSettings(number_format=fmt)
            settings.parse_gerber_value('00001111')


def test_write_format_validation():
    for fmt in (7,5), (5,8), (13,1):
        with pytest.raises(ValueError):
            settings = FileSettings(number_format=fmt)
            settings.write_gerber_value(69.0)





from pathlib import Path

import pytest

from .image_support import ImageDifference

def pytest_assertrepr_compare(op, left, right):
    if isinstance(left, ImageDifference) or isinstance(right, ImageDifference):
        diff = left if isinstance(left, ImageDifference) else right
        return [
            f'Image difference assertion failed.',
            f'    Calculated difference: {diff}',
            f'    Histogram: {diff.histogram}', ]

# store report in node object so tmp_gbr can determine if the test failed.
@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    outcome = yield
    rep = outcome.get_result()
    setattr(item, f'rep_{rep.when}', rep)

fail_dir = Path('gerbonara_test_failures')
def pytest_sessionstart(session):
    if not hasattr(session.config, 'workerinput'): # on worker
        return

    # on coordinator
    for f in chain(fail_dir.glob('*.gbr'), fail_dir.glob('*.png')):
        f.unlink()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2022 Jan Götte <code@jaseg.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Based on https://github.com/tracespace/tracespace
#

import subprocess
from pathlib import Path
import tempfile
import textwrap
import os
from functools import total_ordering
import shutil
import bs4
from contextlib import contextmanager
import hashlib

import numpy as np
from PIL import Image

cachedir = Path(__file__).parent / 'image_cache'
cachedir.mkdir(exist_ok=True)

@total_ordering
class ImageDifference:
    def __init__(self, value, histogram):
        self.value = value
        self.histogram = histogram

    def __float__(self):
        return float(self.value)

    def __eq__(self, other):
        return float(self) == float(other)

    def __lt__(self, other):
        return float(self) < float(other)

    def __str__(self):
        return str(float(self))

@total_ordering
class Histogram:
    def __init__(self, value, size):
        self.value, self.size = value, size

    def __eq__(self, other):
        other = np.array(other)
        other[other == None] = self.value[other == None]
        return (self.value == other).all()

    def __lt__(self, other):
        other = np.array(other)
        other[other == None] = self.value[other == None]
        return (self.value <= other).all()

    def __getitem__(self, index):
        return self.value[index]

    def __str__(self):
        return f'{list(self.value)} size={self.size}'


def run_cargo_cmd(cmd, args, **kwargs):
    if cmd.upper() in os.environ:
        return subprocess.run([os.environ[cmd.upper()], *args], **kwargs)

    try:
        return subprocess.run([cmd, *args], **kwargs)

    except FileNotFoundError:
        return subprocess.run([str(Path.home() / '.cargo' / 'bin' / cmd), *args], **kwargs)

def svg_to_png(in_svg, out_png, dpi=100, bg=None):
    params = f'{dpi}{bg}'.encode()
    digest = hashlib.blake2b(Path(in_svg).read_bytes() + params).hexdigest()
    cachefile = cachedir / f'{digest}.png'

    if not cachefile.is_file():
        bg = 'black' if bg is None else bg
        run_cargo_cmd('resvg', ['--background', bg, '--dpi', str(dpi), in_svg, cachefile], check=True, stdout=subprocess.DEVNULL)

    shutil.copy(cachefile, out_png)

to_gerbv_svg_units = lambda val, unit='mm': val*72 if unit == 'inch' else val/25.4*72

def gerbv_export(in_gbr, out_svg, export_format='svg', origin=(0, 0), size=(6, 6), fg='#ffffff', bg='#000000', override_unit_spec=None):
    params = f'{origin}{size}{fg}{bg}'.encode()
    digest = hashlib.blake2b(Path(in_gbr).read_bytes() + params).hexdigest()
    cachefile = cachedir / f'{digest}.svg'

    if not cachefile.is_file():
        # NOTE: gerbv seems to always export 'clear' polarity apertures as white, irrespective of --foreground, --background
        # and project file color settings.
        # TODO: File issue upstream.
        with tempfile.NamedTemporaryFile('w') as f:
            if override_unit_spec:
                units, zeros, digits = override_unit_spec
                print(f'{Path(in_gbr).name}: overriding excellon unit spec to {units=} {zeros=} {digits=}')
                units = 0 if units == 'inch' else 1
                zeros = {None: 0, 'leading': 1, 'trailing': 2}[zeros]
                unit_spec = textwrap.dedent(f'''(cons 'attribs (list
                        (list 'autodetect 'Boolean 0)
                        (list 'zero_suppression 'Enum {zeros})
                        (list 'units 'Enum {units})
                        (list 'digits 'Integer {digits})
                    ))''')
            else:
                unit_spec = ''

            r, g, b = int(fg[1:3], 16), int(fg[3:5], 16), int(fg[5:], 16)
            color = f"(cons 'color #({r*257} {g*257} {b*257}))"
            f.write(f'''(gerbv-file-version! "2.0A")(define-layer! 0 (cons 'filename "{in_gbr}"){unit_spec}{color})''')
            f.flush()
            if override_unit_spec:
                shutil.copy(f.name, '/tmp/foo.gbv')

            x, y = origin
            w, h = size
            cmd = ['gerbv', '-x', export_format,
                '--border=0',
                f'--origin={x:.6f}x{y:.6f}', f'--window_inch={w:.6f}x{h:.6f}',
                f'--background={bg}',
                f'--foreground={fg}',
                '-o', str(cachefile), '-p', f.name]
            subprocess.run(cmd, check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    shutil.copy(cachefile, out_svg)

@contextmanager
def svg_soup(filename):
    with open(filename, 'r') as f:
        soup = bs4.BeautifulSoup(f.read(), 'xml')

    yield soup

    with open(filename, 'w') as f:
        f.write(str(soup))

def cleanup_gerbv_svg(soup):
    for group in soup.find_all('g'):
        # gerbv uses Cairo's SVG canvas. Cairo's SVG canvas is kind of broken. It has no support for unit
        # handling at all, which means the output files just end up being in pixels at 72 dpi. Further, it
        # seems gerbv's aperture macro rendering interacts poorly with Cairo's SVG export. gerbv renders
        # aperture macros into a new surface, which for some reason gets clipped by Cairo to the given
        # canvas size. This is just wrong, so we just nuke the clip path from these SVG groups here.
        #
        # Apart from being graphically broken, this additionally causes very bad rendering performance.
        del group['clip-path']

def gerber_difference(reference, actual, diff_out=None, svg_transform=None, size=(10,10), ref_unit_spec=None):
    with tempfile.NamedTemporaryFile(suffix='.svg') as act_svg,\
        tempfile.NamedTemporaryFile(suffix='.svg') as ref_svg:

        gerbv_export(reference, ref_svg.name, size=size, export_format='svg', override_unit_spec=ref_unit_spec)
        gerbv_export(actual, act_svg.name, size=size, export_format='svg')

        with svg_soup(ref_svg.name) as soup:
            if svg_transform is not None:
                soup.find('g', attrs={'id': 'surface1'})['transform'] = svg_transform
            cleanup_gerbv_svg(soup)

        with svg_soup(act_svg.name) as soup:
            cleanup_gerbv_svg(soup)

        return svg_difference(ref_svg.name, act_svg.name, diff_out=diff_out)

def gerber_difference_merge(ref1, ref2, actual, diff_out=None, composite_out=None, svg_transform1=None, svg_transform2=None, size=(10,10)):
    with tempfile.NamedTemporaryFile(suffix='.svg') as act_svg,\
        tempfile.NamedTemporaryFile(suffix='.svg') as ref1_svg,\
        tempfile.NamedTemporaryFile(suffix='.svg') as ref2_svg:

        gerbv_export(ref1, ref1_svg.name, size=size, export_format='svg')
        gerbv_export(ref2, ref2_svg.name, size=size, export_format='svg')
        gerbv_export(actual, act_svg.name, size=size, export_format='svg')

        with svg_soup(ref1_svg.name) as soup1:
            if svg_transform1 is not None:
                soup1.find('g', attrs={'id': 'surface1'})['transform'] = svg_transform1
            cleanup_gerbv_svg(soup1)

            with svg_soup(ref2_svg.name) as soup2:
                if svg_transform2 is not None:
                    soup2.find('g', attrs={'id': 'surface1'})['transform'] = svg_transform2
                cleanup_gerbv_svg(soup2)

                defs1 = soup1.find('defs')
                if not defs1:
                    defs1 = soup1.new_tag('defs')
                    soup1.find('svg').insert(0, defs1)

                defs2 = soup2.find('defs')
                if defs2:
                    defs2 = defs2.extract()
                    # explicitly convert .contents into list here and below because else bs4 stumbles over itself
                    # iterating because we modify the tree in the loop body.
                    for c in list(defs2.contents):
                        if hasattr(c, 'attrs'):
                            c['id'] = 'gn-merge-b-' + c.attrs.get('id', str(id(c)))
                        defs1.append(c)

                for use in soup2.find_all('use', recursive=True):
                    if (href := use.get('xlink:href', '')).startswith('#'):
                        use['xlink:href'] = f'#gn-merge-b-{href[1:]}'

                svg1 = soup1.find('svg')
                for c in list(soup2.find('svg').contents):
                    if hasattr(c, 'attrs'):
                        c['id'] = 'gn-merge-b-' + c.attrs.get('id', str(id(c)))
                    svg1.append(c)

        if composite_out:
            shutil.copyfile(ref1_svg.name, composite_out)

        with svg_soup(act_svg.name) as soup:
            cleanup_gerbv_svg(soup)

        return svg_difference(ref1_svg.name, act_svg.name, diff_out=diff_out)

def svg_difference(reference, actual, diff_out=None, background=None):
    with tempfile.NamedTemporaryFile(suffix='-ref.png') as ref_png,\
        tempfile.NamedTemporaryFile(suffix='-act.png') as act_png:

        svg_to_png(reference, ref_png.name, bg=background)
        svg_to_png(actual, act_png.name, bg=background)

        return image_difference(ref_png.name, act_png.name, diff_out=diff_out)

def image_difference(reference, actual, diff_out=None):
    ref = np.array(Image.open(reference)).astype(float)
    out = np.array(Image.open(actual)).astype(float)

    ref, out = ref.mean(axis=2), out.mean(axis=2) # convert to grayscale
    # TODO blur images here before comparison to mitigate aliasing issue
    delta = np.abs(out - ref).astype(float) / 255
    if diff_out:
        Image.fromarray((delta*255).astype(np.uint8), mode='L').save(diff_out)

    hist, _bins = np.histogram(delta, bins=10, range=(0, 1))
    return (ImageDifference(delta.mean(), hist),
            ImageDifference(delta.max(), hist),
            Histogram(hist, out.size))





#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2022 Jan Götte <code@jaseg.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import math

import pytest
from scipy.spatial import KDTree

from ..excellon import ExcellonFile
from ..rs274x import GerberFile
from ..cam import FileSettings
from ..graphic_objects import Flash

from .image_support import *
from .utils import *
from ..utils import Inch, MM

REFERENCE_FILES = {
        'easyeda/Gerber_Drill_NPTH.DRL': (None, None),
        'easyeda/Gerber_Drill_PTH.DRL': (None, 'easyeda/Gerber_TopLayer.GTL'),
        # Altium uses an excellon format specification format that gerbv doesn't understand, so we have to fix that.
        'altium-composite-drill/NC Drill/LimeSDR-QPCIe_1v2-SlotHoles.TXT': (('mm', 'trailing', 4), None),
        'altium-composite-drill/NC Drill/LimeSDR-QPCIe_1v2-RoundHoles.TXT': (('mm', 'trailing', 4), 'altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.GTL'),
        'pcb-rnd/power-art.xln': (None, 'pcb-rnd/power-art.gtl'),
        'siemens/80101_0125_F200_ThruHoleNonPlated.ncd': (None, None),
        'siemens/80101_0125_F200_ThruHolePlated.ncd': (None, 'siemens/80101_0125_F200_L01_Top.gdo'),
        'siemens/80101_0125_F200_ContourPlated.ncd': (None, None),
        'Target3001/IRNASIoTbank1.2.Drill': (None, 'Target3001/IRNASIoTbank1.2.Top'),
        'altium-old-composite-drill.txt': (None, None),
        'fritzing/combined.txt': (None, 'fritzing/combined.gtl'),
        'ncdrill.DRD': (None, None),
        'upverter/design_export.drl': (None, 'upverter/design_export.gtl'),
        'diptrace/mainboard.drl': (None, 'diptrace/mainboard_Top.gbr'),
        'diptrace/panel.drl': (None, None),
        'diptrace/keyboard.drl': (None, 'diptrace/keyboard_Bottom.gbr'),
        'zuken-emulated/Drill/8seg_Driver__routed_Drill_thru_plt.fdr/8seg_Driver__routed_Drill_thru_plt.fdr': (('inch', 'trailing', 4), 'zuken-emulated/Gerber/Conductive-1.fph'),
        'zuken-emulated/Drill/8seg_Driver__routed_Drill_thru_nplt.fdr': (('inch', 'trailing', 4), None),
        }

@filter_syntax_warnings
@pytest.mark.parametrize('reference', list(REFERENCE_FILES.items()), indirect=True)
def test_round_trip(reference, tmpfile):
    reference, (unit_spec, _) = reference
    tmp = tmpfile('Output excellon', '.drl')
    print('unit spec', unit_spec)

    f = ExcellonFile.open(reference)
    f.save(tmp)

    if reference.name == '80101_0125_F200_ContourPlated.ncd':
        # gerbv does not support routed slots in excellon files at all and renders garbage for the reference file here
        # due to its use of bare coordinates for routed slots. Thus, we skip this test (for now).
        return

    mean, _max, hist = gerber_difference(reference, tmp, diff_out=tmpfile('Difference', '.png'), ref_unit_spec=unit_spec)
    assert mean < 5e-5
    assert hist[9] == 0
    assert hist[3:].sum() < 5e-5*hist.size

@filter_syntax_warnings
@pytest.mark.parametrize('reference', list(REFERENCE_FILES.items()), indirect=True)
def test_first_level_idempotence_svg(reference, tmpfile):
    reference, (unit_spec, _) = reference
    tmp = tmpfile('Output excellon', '.drl')
    ref_svg = tmpfile('Reference SVG render', '.svg')
    out_svg = tmpfile('Output SVG render', '.svg')
    print('unit spec', unit_spec)

    a = ExcellonFile.open(reference)
    a.save(tmp)
    b = ExcellonFile.open(tmp)

    ref_svg.write_text(str(a.to_svg(fg='black', bg='white')))
    out_svg.write_text(str(b.to_svg(fg='black', bg='white')))

    mean, _max, hist = svg_difference(ref_svg, out_svg, diff_out=tmpfile('Difference', '.png'), background='white')
    assert mean < 5e-5
    assert hist[9] == 0
    assert hist[3:].sum() < 5e-5*hist.size

@filter_syntax_warnings
@pytest.mark.parametrize('reference', list(REFERENCE_FILES.items()), indirect=True)
def test_idempotence(reference, tmpfile):
    reference, (unit_spec, _) = reference

    if reference.name == '80101_0125_F200_ContourPlated.ncd':
        # this file contains a duplicate tool definition that we optimize out on our second pass.
        # TODO see whether we can change things so we optimize this out on the first pass already. I'm not sure what
        # went wrong there.
        pytest.skip()

    tmp_1 = tmpfile('First generation output', '.drl')
    tmp_2 = tmpfile('Second generation output', '.drl')

    f1 = ExcellonFile.open(reference)
    f1.save(tmp_1)
    print(f'{f1.import_settings=}')
    f2 = ExcellonFile.open(tmp_1)
    f2.save(tmp_2)
    print(f'{f2.import_settings=}')

    assert tmp_1.read_text() == tmp_2.read_text()

@filter_syntax_warnings
@pytest.mark.parametrize('reference', list(REFERENCE_FILES.items()), indirect=True)
def test_gerber_alignment(reference, tmpfile, print_on_error):
    reference, (unit_spec, gerber) = reference
    tmp = tmpfile('Output excellon', '.drl')

    if gerber is None:
        pytest.skip()

    excf = ExcellonFile.open(reference)
    gerf_path = reference_path(gerber)
    print_on_error('Reference gerber file:', gerf_path)
    gerf = GerberFile.open(gerf_path)
    print('bounds excellon:', excf.bounding_box(MM))
    print('bounds gerber:', gerf.bounding_box(MM))
    excf.save(tmp)

    flash_coords = []
    for obj in gerf.objects:
        if isinstance(obj, Flash):
            x, y = obj.unit.convert_to(MM, obj.x), obj.unit.convert_to(MM, obj.y)
            flash_coords.append((x, y))

    tree = KDTree(flash_coords, copy_data=True)

    tolerance = 0.05 # mm
    matches, total = 0, 0
    for obj in excf.objects:
        if isinstance(obj, Flash):
            if obj.plated in (True, None):
                total += 1
                x, y = obj.unit.convert_to(MM, obj.x), obj.unit.convert_to(MM, obj.y)
                if tree.query_ball_point((x, y), r=tolerance):
                    matches += 1

    # Some PCB tools, notably easyeda, are dumb and export certain pads as regions, not apertures. Thus, we have to
    # tolerate some non-matches.
    assert matches > 10
    assert matches/total > 0.5

@filter_syntax_warnings
def test_syntax_error():
    ref = reference_path('test_syntax_error.exc')
    with pytest.raises(SyntaxError) as exc_info:
        ExcellonFile.open(ref)

    assert 'test_syntax_error.exc' in exc_info.value.msg
    assert '12' in exc_info.value.msg # lineno




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2022 Jan Götte <code@jaseg.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Based on https://github.com/tracespace/tracespace
#

import math

from PIL import Image
import pytest

from ..rs274x import GerberFile
from ..cam import FileSettings

from .image_support import *
from .utils import *

REFERENCE_FILES = [ l.strip() for l in '''
    board_outline.GKO
    example_outline_with_arcs.gbr
    example_two_square_boxes.gbr
    example_coincident_hole.gbr
    example_cutin.gbr
    example_cutin_multiple.gbr
    example_flash_circle.gbr
    example_flash_obround.gbr
    example_flash_polygon.gbr
    example_flash_rectangle.gbr
    example_fully_coincident.gbr
    example_guess_by_content.g0
    example_holes_dont_clear.gbr
    example_level_holes.gbr
    example_not_overlapping_contour.gbr
    example_not_overlapping_touching.gbr
    example_overlapping_contour.gbr
    example_overlapping_touching.gbr
    example_simple_contour.gbr
    example_single_contour_1.gbr
    example_single_contour_2.gbr
    example_single_contour_3.gbr
    example_am_exposure_modifier.gbr
    bottom_copper.GBL
    bottom_mask.GBS
    bottom_silk.GBO
    eagle_files/copper_bottom_l4.gbr
    eagle_files/copper_inner_l2.gbr
    eagle_files/copper_inner_l3.gbr
    eagle_files/copper_top_l1.gbr
    eagle_files/profile.gbr
    eagle_files/silkscreen_bottom.gbr
    eagle_files/silkscreen_top.gbr
    eagle_files/soldermask_bottom.gbr
    eagle_files/soldermask_top.gbr
    eagle_files/solderpaste_bottom.gbr
    eagle_files/solderpaste_top.gbr
    multiline_read.ger
    test_fine_lines_x.gbr
    test_fine_lines_y.gbr
    top_copper.GTL
    top_mask.GTS
    top_silk.GTO
    open_outline_altium.gbr
    easyeda/Gerber_TopSolderMaskLayer.GTS
    easyeda/Gerber_TopSilkLayer.GTO
    easyeda/Gerber_BottomSolderMaskLayer.GBS
    easyeda/Gerber_BoardOutline.GKO
    easyeda/Gerber_TopLayer.GTL
    easyeda/Gerber_BottomLayer.GBL
    easyeda/Gerber_TopPasteMaskLayer.GTP
    allegro-2/MinnowMax_RevA1_GAF_Gerber/MinnowMax_lyr2.art
    allegro-2/MinnowMax_RevA1_GAF_Gerber/MinnowMax_lyr3.art
    allegro-2/MinnowMax_RevA1_GAF_Gerber/MinnowMax_fab.art
    allegro-2/MinnowMax_RevA1_GAF_Gerber/MinnowMax_lyr10_GAF.art
    allegro-2/MinnowMax_RevA1_GAF_Gerber/MinnowMax_lyr7.art
    allegro-2/MinnowMax_RevA1_GAF_Gerber/MinnowMax_sps.art
    allegro-2/MinnowMax_RevA1_GAF_Gerber/MinnowMax_lyr6.art
    allegro-2/MinnowMax_RevA1_GAF_Gerber/MinnowMax_lyr1_GAF.art
    allegro-2/MinnowMax_RevA1_GAF_Gerber/MinnowMax_assy.art
    allegro-2/MinnowMax_RevA1_GAF_Gerber/MinnowMax_smc_GAF.art
    allegro-2/MinnowMax_RevA1_GAF_Gerber/MinnowMax_lyr4.art
    allegro-2/MinnowMax_RevA1_GAF_Gerber/MinnowMax_lyr5.art
    allegro-2/MinnowMax_RevA1_GAF_Gerber/MinnowMax_bslk.art
    allegro-2/MinnowMax_RevA1_GAF_Gerber/MinnowMax_spc.art
    allegro-2/MinnowMax_RevA1_GAF_Gerber/MinnowMax_tslk_GAF.art
    allegro-2/MinnowMax_RevA1_GAF_Gerber/MinnowMax_lyr8.art
    allegro-2/MinnowMax_RevA1_GAF_Gerber/MinnowMax_sms_GAF.art
    allegro-2/MinnowMax_RevA1_GAF_Gerber/MinnowMax_lyr9.art
    eagle-newer/solderpaste_bottom.gbr
    eagle-newer/silkscreen_bottom.gbr
    eagle-newer/profile.gbr
    eagle-newer/copper_bottom.gbr
    eagle-newer/soldermask_top.gbr
    eagle-newer/solderpaste_top.gbr
    eagle-newer/soldermask_bottom.gbr
    eagle-newer/silkscreen_top.gbr
    eagle-newer/copper_top.gbr
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.G4
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.G9
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.GBL
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.GTO
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.G11
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.G1
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.GBP
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.G2
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.GM15
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.GTS
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.G6
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.G7
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.G3
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.GPB
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.GM1
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.G12
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.GBS
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.GTL
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.G10
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.GM14
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.G5
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.GTP
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.GBO
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.G8
    altium-composite-drill/Gerber/LimeSDR-QPCIe_1v2.GPT
    geda/driver.topmask.gbr
    geda/controller.top.gbr
    geda/controller.bottom.gbr
    geda/driver.bottommask.gbr
    geda/driver.top.gbr
    geda/driver.bottom.gbr
    geda/controller.topsilk.gbr
    geda/controller.fab.gbr
    geda/driver.topsilk.gbr
    geda/controller.group3.gbr
    geda/controller.topmask.gbr
    geda/driver.group5.gbr
    geda/controller.bottommask.gbr
    geda/driver.fab.gbr
    pcb-rnd/power-art.gko
    pcb-rnd/power-art.ast
    pcb-rnd/power-art.gtl
    pcb-rnd/power-art.lht
    pcb-rnd/power-art.gto
    pcb-rnd/power-art.gtp
    pcb-rnd/power-art.asb
    pcb-rnd/power-art.gbp
    pcb-rnd/power-art.gbs
    pcb-rnd/power-art.gbl
    pcb-rnd/power-art.fab
    pcb-rnd/power-art.gbo
    pcb-rnd/power-art.gts
    siemens/80101_0125_F200_L04.gdo
    siemens/80101_0125_F200_L12_Bottom.gdo
    siemens/80101_0125_F200_L11.gdo
    siemens/80101_0125_F200_L10.gdo
    siemens/80101_0125_F200_SolderPasteTop.gdo
    siemens/80101_0125_F200_SoldermaskTop.gdo
    siemens/80101_0125_F200_L06.gdo
    siemens/80101_0125_F200_L02.gdo
    siemens/80101_0125_F200_SilkscreenBottom.gdo
    siemens/80101_0125_F200_SoldermaskBottom.gdo
    siemens/80101_0125_F200_SolderPasteBottom.gdo
    siemens/80101_0125_F200_L03.gdo
    siemens/80101_0125_F200_L01_Top.gdo
    Target3001/IRNASIoTbank1.2.Bot
    Target3001/IRNASIoTbank1.2.Outline
    Target3001/IRNASIoTbank1.2.PasteBot
    Target3001/IRNASIoTbank1.2.PasteTop
    Target3001/IRNASIoTbank1.2.PosiBot
    Target3001/IRNASIoTbank1.2.PosiTop
    Target3001/IRNASIoTbank1.2.StopBot
    Target3001/IRNASIoTbank1.2.StopTop
    Target3001/IRNASIoTbank1.2.Top
    kicad-older/chibi_2024-Edge.Cuts.gbr
    kicad-older/chibi_2024-F.SilkS.gbr
    kicad-older/chibi_2024-B.Paste.gbr
    kicad-older/chibi_2024-B.Cu.gbr
    kicad-older/chibi_2024-F.Mask.gbr
    kicad-older/chibi_2024-B.Mask.gbr
    kicad-older/chibi_2024-F.Paste.gbr
    kicad-older/chibi_2024-B.SilkS.gbr
    kicad-older/chibi_2024-F.Cu.gbr
    fritzing/combined.gbs
    fritzing/combined.gm1
    fritzing/combined.gbl
    fritzing/combined.gbo
    fritzing/combined.GKO
    fritzing/combined.gtl
    fritzing/combined.gts
    fritzing/combined.gto
    siemens-2/Gerber/SoldermaskTop.gdo
    siemens-2/Gerber/EtchLayerTop.gdo
    siemens-2/Gerber/DrillDrawingThrough.gdo
    siemens-2/Gerber/SoldermaskBottom.gdo
    siemens-2/Gerber/SolderPasteBottom.gdo
    siemens-2/Gerber/SolderPasteTop.gdo
    siemens-2/Gerber/EtchLayerBottom.gdo
    siemens-2/Gerber/GerberPlot.gpf
    siemens-2/Gerber/BoardOutlline.gdo
    upverter/design_export.gko
    upverter/design_export.gtl
    upverter/design_export.gbp
    upverter/design_export.gtp
    upverter/design_export.gbl
    upverter/design_export.gto
    upverter/design_export.gbs
    upverter/design_export.gts
    upverter/design_export.gbo
    eagle_files/solderpaste_bottom.gbr
    eagle_files/silkscreen_bottom.gbr
    eagle_files/profile.gbr
    eagle_files/copper_inner_l2.gbr
    eagle_files/copper_top_l1.gbr
    eagle_files/soldermask_top.gbr
    eagle_files/copper_inner_l3.gbr
    eagle_files/solderpaste_top.gbr
    eagle_files/soldermask_bottom.gbr
    eagle_files/copper_bottom_l4.gbr
    eagle_files/silkscreen_top.gbr
    diptrace/panel_BoardOutline.gbr
    diptrace/keyboard_BottomSilk.gbr
    diptrace/keyboard_Bottom.gbr
    diptrace/mainboard_Top.gbr
    diptrace/mainboard_TopMask.gbr
    diptrace/mainboard_BoardOutline.gbr
    diptrace/mainboard_Bottom.gbr
    diptrace/mainboard_BottomMask.gbr
    diptrace/keyboard_BottomMask.gbr
    diptrace/panel_Bottom.gbr
    diptrace/keyboard_BoardOutline.gbr
    diptrace/panel_BottomSilk.gbr
    diptrace/panel_BottomMask.gbr
    diptrace/mainboard_TopSilk.gbr
    zuken-emulated/Gerber/MetalMask-A.fph
    zuken-emulated/Gerber/MetalMask-B.fph
    zuken-emulated/Gerber/Symbol-A.fph
    zuken-emulated/Gerber/Symbol-B.fph
    zuken-emulated/Gerber/Resist-A.fph
    zuken-emulated/Gerber/Resist-B.fph
    zuken-emulated/Gerber/Conductive-1.fph
    zuken-emulated/Gerber/Conductive-2.fph
'''.splitlines() if l ]

MIN_REFERENCE_FILES = [
    'example_two_square_boxes.gbr',
    'example_outline_with_arcs.gbr',
    'example_flash_circle.gbr',
    'example_flash_polygon.gbr',
    'example_flash_rectangle.gbr',
    'example_simple_contour.gbr',
    'example_am_exposure_modifier.gbr',
    'bottom_copper.GBL',
    'bottom_silk.GBO',
    'eagle_files/copper_bottom_l4.gbr'
    ]

HAS_ZERO_SIZE_APERTURES = [
    'bottom_copper.GBL',
    'bottom_silk.GBO',
    'top_copper.GTL',
    'top_silk.GTO',
    'board_outline.GKO',
    'silkscreen_top.gbr',
    'combined.GKO',
    'combined.gto',
    'EtchLayerTop.gdo',
    'EtchLayerBottom.gdo',
    'BoardOutlline.gdo',
    ]


@filter_syntax_warnings
@pytest.mark.parametrize('reference', REFERENCE_FILES, indirect=True)
def test_round_trip(reference, tmpfile):
    tmp_gbr = tmpfile('Output gerber', '.gbr')

    GerberFile.open(reference).save(tmp_gbr)

    mean, _max, hist = gerber_difference(reference, tmp_gbr, diff_out=tmpfile('Difference', '.png'))
    assert mean < 5e-5
    assert hist[9] == 0
    assert hist[3:].sum() < 5e-5*hist.size

@filter_syntax_warnings
@pytest.mark.parametrize('reference', REFERENCE_FILES, indirect=True)
def test_idempotence(reference, tmpfile):
    tmp_gbr_1 = tmpfile('First generation output', '.gbr')
    tmp_gbr_2 = tmpfile('Second generation output', '.gbr')

    GerberFile.open(reference).save(tmp_gbr_1)
    GerberFile.open(tmp_gbr_1).save(tmp_gbr_2)
    assert tmp_gbr_1.read_text() == tmp_gbr_2.read_text()


TEST_ANGLES = [90, 180, 270, 30, 1.5, 10, 360, 1024, -30, -90]
TEST_OFFSETS = [(0, 0), (100, 0), (0, 100), (2, 0), (10, 100)]

@filter_syntax_warnings
@pytest.mark.parametrize('reference', MIN_REFERENCE_FILES, indirect=True)
@pytest.mark.parametrize('angle', TEST_ANGLES)
def test_rotation(reference, angle, tmpfile):
    if 'flash_rectangle' in str(reference) and angle == 1024:
        # gerbv's rendering of this is broken, the hole is missing.
        pytest.skip()

    tmp_gbr = tmpfile('Output gerber', '.gbr')

    f = GerberFile.open(reference)
    f.rotate(math.radians(angle))
    f.save(tmp_gbr)

    cx, cy = 0, to_gerbv_svg_units(10, unit='inch')
    mean, _max, hist = gerber_difference(reference, tmp_gbr, diff_out=tmpfile('Difference', '.png'),
            svg_transform=f'rotate({angle} {cx} {cy})')
    assert mean < 1e-3 # relax mean criterion compared to above.
    assert hist[9] == 0

@filter_syntax_warnings
@pytest.mark.parametrize('reference', MIN_REFERENCE_FILES, indirect=True)
@pytest.mark.parametrize('angle', TEST_ANGLES)
@pytest.mark.parametrize('center', [(0, 0), (-10, -10), (10, 10), (10, 0), (0, -10), (-10, 10), (10, 20)])
def test_rotation_center(reference, angle, center, tmpfile):
    if 'flash_rectangle' in str(reference) and angle in (30, 1024):
        # gerbv's rendering of this is broken, the hole is missing.
        pytest.skip()

    tmp_gbr = tmpfile('Output gerber', '.gbr')

    f = GerberFile.open(reference)
    f.rotate(math.radians(angle), center=center)
    f.save(tmp_gbr)

    # calculate circle center in SVG coordinates 
    size = (10, 10) # inches
    cx, cy = to_gerbv_svg_units(center[0]), to_gerbv_svg_units(size[1], 'inch')-to_gerbv_svg_units(center[1], 'mm')
    mean, _max, hist = gerber_difference(reference, tmp_gbr, diff_out=tmpfile('Difference', '.png'),
            svg_transform=f'rotate({angle} {cx} {cy})',
            size=size)
    assert mean < 1e-3
    assert hist[9] < 50
    assert hist[3:].sum() < 1e-3*hist.size

@filter_syntax_warnings
@pytest.mark.parametrize('reference', MIN_REFERENCE_FILES, indirect=True)
@pytest.mark.parametrize('offset', TEST_OFFSETS)
def test_offset(reference, offset, tmpfile):
    tmp_gbr = tmpfile('Output gerber', '.gbr')

    f = GerberFile.open(reference)
    f.offset(*offset)
    f.save(tmp_gbr, settings=FileSettings(unit=f.unit, number_format=(4,7)))

    # flip y offset since svg's y axis is flipped compared to that of gerber
    dx, dy = to_gerbv_svg_units(offset[0]), -to_gerbv_svg_units(offset[1])
    mean, _max, hist = gerber_difference(reference, tmp_gbr, diff_out=tmpfile('Difference', '.png'),
            svg_transform=f'translate({dx} {dy})')
    assert mean < 1e-4
    assert hist[9] == 0

@filter_syntax_warnings
@pytest.mark.parametrize('reference', MIN_REFERENCE_FILES, indirect=True)
@pytest.mark.parametrize('angle', TEST_ANGLES)
@pytest.mark.parametrize('center', [(0, 0), (10, 0), (0, -10), (10, 20)])
@pytest.mark.parametrize('offset', [(0, 0), (100, 0), (0, 100), (100, 100), (100, 10)])
def test_combined(reference, angle, center, offset, tmpfile):
    if 'flash_rectangle' in str(reference) and angle in (30, 1024):
        # gerbv's rendering of this is broken, the hole is missing.
        pytest.skip()

    tmp_gbr = tmpfile('Output gerber', '.gbr')

    f = GerberFile.open(reference)
    f.rotate(math.radians(angle), center=center)
    f.offset(*offset)
    f.save(tmp_gbr, settings=FileSettings(unit=f.unit, number_format=(4,7)))

    size = (10, 10) # inches
    cx, cy = to_gerbv_svg_units(center[0]), to_gerbv_svg_units(size[1], 'inch')-to_gerbv_svg_units(center[1], 'mm')
    dx, dy = to_gerbv_svg_units(offset[0]), -to_gerbv_svg_units(offset[1])
    mean, _max, hist = gerber_difference(reference, tmp_gbr, diff_out=tmpfile('Difference', '.png'),
            svg_transform=f'translate({dx} {dy}) rotate({angle} {cx} {cy})',
            size=size)
    assert mean < 1e-3
    assert hist[9] < 100
    assert hist[3:].sum() < 1e-3*hist.size

@filter_syntax_warnings
@pytest.mark.parametrize('file_a', MIN_REFERENCE_FILES)
@pytest.mark.parametrize('file_b', [
    'example_two_square_boxes.gbr',
    'example_outline_with_arcs.gbr',
    'example_am_exposure_modifier.gbr',
    'bottom_silk.GBO',
    'eagle_files/copper_bottom_l4.gbr', ])
@pytest.mark.parametrize('angle', [0, 10, 90])
@pytest.mark.parametrize('offset', [(0, 0, 0, 0), (100, 0, 0, 0), (0, 0, 0, 100), (100, 0, 0, 100)])
def test_compositing(file_a, file_b, angle, offset, tmpfile, print_on_error):

    # TODO bottom_silk.GBO renders incorrectly with gerbv: the outline does not exist in svg. In GUI, the logo only
    # renders at very high magnification. Skip, and once we have our own SVG export maybe use that instead. Or just use
    # KiCAD's gerbview.
    # TODO check if this and the issue with aperture holes not rendering in test_combined actually are bugs in gerbv
    # and fix/report upstream.
    if file_a == 'bottom_silk.GBO' or file_b == 'bottom_silk.GBO':
        pytest.skip()

    ref_a = reference_path(file_a)
    print_on_error('Reference file a:', ref_a)
    ref_b = reference_path(file_b)
    print_on_error('Reference file b:', ref_b)

    ax, ay, bx, by = offset
    grb_a = GerberFile.open(ref_a)
    grb_a.rotate(math.radians(angle))
    grb_a.offset(ax, ay)

    grb_b = GerberFile.open(ref_b)
    grb_b.offset(bx, by)

    grb_a.merge(grb_b)
    tmp_gbr = tmpfile('Output gerber', '.gbr')
    grb_a.save(tmp_gbr, settings=FileSettings(unit=grb_a.unit, number_format=(4,7)))

    size = (10, 10) # inches
    ax, ay = to_gerbv_svg_units(ax), -to_gerbv_svg_units(ay)
    bx, by = to_gerbv_svg_units(bx), -to_gerbv_svg_units(by)
    # note that we have to specify cx, cy even if we rotate around the origin since gerber's origin lies at (x=0
    # y=+document size) in SVG's coordinate space because svg's y axis is flipped compared to gerber's.
    cx, cy = 0, to_gerbv_svg_units(size[1], 'inch')
    mean, _max, hist = gerber_difference_merge(ref_a, ref_b, tmp_gbr,
            composite_out=tmpfile('Composite', '.svg'), diff_out=tmpfile('Difference', '.png'),
            svg_transform1=f'translate({ax} {ay}) rotate({angle} {cx} {cy})',
            svg_transform2=f'translate({bx} {by})',
            size=size)
    assert mean < 1e-3
    assert hist[9] < 100
    assert hist[3:].sum() < 1e-3*hist.size

@filter_syntax_warnings
@pytest.mark.parametrize('reference', REFERENCE_FILES, indirect=True)
def test_svg_export(reference, tmpfile):
    if reference.name in ('silkscreen_bottom.gbr', 'silkscreen_top.gbr', 'top_silk.GTO'):
        # Some weird svg rendering artifact. Might be caused by mismatching svg units between gerbv and us. Result looks
        # fine though.
        pytest.skip()

    if reference.name == 'MinnowMax_assy.art':
        # This leads to worst-case performance in resvg, this testcase takes over 1h to finish. So skip.
        pytest.skip()

    grb = GerberFile.open(reference)

    bounds = (0.0, 0.0), (6.0, 6.0) # bottom left, top right

    out_svg = tmpfile('Output', '.svg')
    with open(out_svg, 'w') as f:
        f.write(str(grb.to_svg(force_bounds=bounds, arg_unit='inch', fg='black', bg='white')))

    # NOTE: Instead of having gerbv directly export a PNG, we ask gerbv to output SVG which we then rasterize using
    # resvg. We have to do this since gerbv's built-in cairo-based PNG export has severe aliasing issues. In contrast,
    # using resvg for both allows an apples-to-apples comparison of both results.
    ref_svg = tmpfile('Reference export', '.svg')
    ref_png = tmpfile('Reference render', '.png')
    gerbv_export(reference, ref_svg, origin=bounds[0], size=bounds[1], fg='#000000', bg='#ffffff')
    svg_to_png(ref_svg, ref_png, dpi=300, bg='white')

    out_png = tmpfile('Output render', '.png')
    svg_to_png(out_svg, out_png, dpi=300, bg='white')

    if reference.name in HAS_ZERO_SIZE_APERTURES:
        # gerbv does not render these correctly.
        return

    mean, _max, hist = image_difference(ref_png, out_png, diff_out=tmpfile('Difference', '.png'))
    assert hist[9] < 1
    if 'Minnow' in reference.name or 'LimeSDR' in reference.name or '80101_0125_F200' in reference.name:
        # This is a dense design with lots of traces, leading to lots of aliasing artifacts.
        assert mean < 10e-3
        assert hist[4:].sum() < 1e-2*hist.size
    else:
        assert mean < 1.2e-3
        assert hist[3:].sum() < 1e-3*hist.size

# FIXME test svg margin, bounding box computation

@filter_syntax_warnings
@pytest.mark.parametrize('reference', REFERENCE_FILES, indirect=True)
def test_bounding_box(reference, tmpfile):
    if reference.name == 'MinnowMax_assy.art':
        # This leads to worst-case performance in resvg, this testcase takes over 1h to finish. So skip.
        pytest.skip()
    # skip this check on files that contain lines with a zero-size aperture at the board edge
    if any(reference.match(f'*/{f}') for f in HAS_ZERO_SIZE_APERTURES):
        pytest.skip()

    margin = 1.0 # inch
    dpi = 200
    margin_px = int(dpi*margin) # intentionally round down to avoid aliasing artifacts

    grb = GerberFile.open(reference)

    if grb.is_empty:
        pytest.skip()

    out_svg = tmpfile('Output', '.svg')
    with open(out_svg, 'w') as f:
        f.write(str(grb.to_svg(margin=margin, arg_unit='inch', fg='white', bg='black')))

    out_png = tmpfile('Render', '.png')
    svg_to_png(out_svg, out_png, dpi=dpi)

    img = np.array(Image.open(out_png))
    img = img[:, :, :3].mean(axis=2) # drop alpha and convert to grayscale
    img = np.round(img).astype(int) # convert to int
    assert (img > 0).any() # there must be some content, none of the test gerbers are completely empty.
    cols = img.sum(axis=1)
    rows = img.sum(axis=0)
    col_prefix, col_suffix = np.argmax(cols > 0), np.argmax(cols[::-1] > 0)
    row_prefix, row_suffix = np.argmax(rows > 0), np.argmax(rows[::-1] > 0)
    print('cols:', col_prefix, col_suffix)
    print('rows:', row_prefix, row_suffix)

    # Check that all margins are completely black and that the content touches the margins. Allow for some tolerance to
    # allow for antialiasing artifacts and for things like very thin features.
    assert margin_px-2 <= col_prefix <= margin_px+2
    assert margin_px-2 <= col_suffix <= margin_px+2
    assert margin_px-2 <= row_prefix <= margin_px+2
    assert margin_px-2 <= row_suffix <= margin_px+2

@filter_syntax_warnings
def test_syntax_error():
    ref = reference_path('test_syntax_error.gbr')
    with pytest.raises(SyntaxError) as exc_info:
        GerberFile.open(ref)

    assert 'test_syntax_error.gbr' in exc_info.value.msg
    assert '7' in exc_info.value.msg # lineno




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2022 Jan Götte <code@jaseg.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
import functools
import tempfile
import re
import shutil
from contextlib import contextmanager
from pathlib import Path

from PIL import Image
import pytest

fail_dir = Path('gerbonara_test_failures')
reference_path = lambda reference: Path(__file__).parent / 'resources' / str(reference)
to_gerbv_svg_units = lambda val, unit='mm': val*72 if unit == 'inch' else val/25.4*72

def path_test_name(request):
    """ Create a slug suitable for use in file names from the test's nodeid """
    module, _, test_name = request.node.nodeid.rpartition('::')
    _test, _, test_name = test_name.partition('_')
    test_name, _, _ext = test_name.partition('.')
    return re.sub(r'[^\w\d]', '_', test_name)

@pytest.fixture
def print_on_error(request):
    messages = []

    def register_print(*args, sep=' ', end='\n'):
        nonlocal messages
        messages.append(sep.join(str(arg) for arg in args) + end)

    yield register_print

    if request.node.rep_call.failed:
        for msg in messages:
            print(msg, end='')

@pytest.fixture
def tmpfile(request):
    registered = []

    def register_tempfile(name, suffix):
        nonlocal registered
        f = tempfile.NamedTemporaryFile(suffix=suffix)
        registered.append((name, suffix, f))
        return Path(f.name)

    yield register_tempfile

    if request.node.rep_call.failed:
        fail_dir.mkdir(exist_ok=True)
        test_name = path_test_name(request)
        for name, suffix, tmp in registered:
            slug = re.sub(r'[^\w\d]+', '_', name.lower())
            perm_path = fail_dir / f'failure_{test_name}_{slug}{suffix}'
            shutil.copy(tmp.name, perm_path)
            print(f'{name} saved to {perm_path}')

    for _name, _suffix, tmp in registered:
        tmp.close()

@pytest.fixture
def reference(request, print_on_error):
    ref = request.param
    if isinstance(ref, tuple):
        ref, args = ref
        ref = reference_path(ref)
        yield ref, args

    else:
        ref = reference_path(request.param)
        yield ref

    print_on_error(f'Reference file: {ref}')

def filter_syntax_warnings(fun):
    a = pytest.mark.filterwarnings('ignore:.*Deprecated.*statement found.*:DeprecationWarning')
    b = pytest.mark.filterwarnings('ignore::SyntaxWarning')
    return a(b(fun))





#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2015 Hamilton Kibbe <ham@hamiltonkib.be>
# Copyright 2022 Jan Götte <code@jaseg.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from ..ipc356 import *
from ..cam import FileSettings

from .utils import *
from ..utils import Inch, MM


REFERENCE_FILES = [
        # TODO add more test files from github
        'allegro-2/MinnowMax_RevA1_IPC356A.ipc',
        'allegro/08_057494d-ipc356.ipc',
        'ipc-d-356.ipc',
        ]


@filter_syntax_warnings
@pytest.mark.parametrize('reference', REFERENCE_FILES, indirect=True)
def test_read(reference):
    netlist = Netlist.open(reference)
    assert netlist
    assert netlist.test_records

@filter_syntax_warnings
@pytest.mark.parametrize('reference', REFERENCE_FILES, indirect=True)
def test_idempotence(reference, tmpfile):
    tmp_1 = tmpfile('First generation output', '.ipc')
    tmp_2 = tmpfile('Second generation output', '.ipc')

    a = Netlist.open(reference)
    a.save(tmp_1)
    b = Netlist.open(tmp_1)
    b.save(tmp_2)

    print(f'{a.outlines=}')
    print(f'{b.outlines=}')

    res = tmp_1.read_text() == tmp_2.read_text()
    # Confuse pytest so it doesn't try to print out a diff. pytest's potato diff algorithm is wayyyy to slow and would
    # hang for several minutes.
    assert res

@filter_syntax_warnings
@pytest.mark.parametrize('reference', REFERENCE_FILES, indirect=True)
def test_bells_and_whistles(reference):
    netlist = Netlist.open(reference)
    netlist.net_names()
    netlist.vias()
    netlist.reference_designators()
    netlist.records_by_reference('C1')
    netlist.records_by_net_name('n001')
    netlist.conductors_by_net_name('n001')
    netlist.conductors_by_layer(0)

@filter_syntax_warnings
@pytest.mark.parametrize('a', REFERENCE_FILES)
@pytest.mark.parametrize('b', REFERENCE_FILES)
def test_merge(a, b):
    a, b = reference_path(a), reference_path(b)
    print('File A:', a)
    print('File B:', a)

    a = Netlist.open(a)
    b = Netlist.open(b)
    a.merge(b, our_prefix='A')
    # FIXME asserts

def test_record_semantics():
    settings = FileSettings()
    r = TestRecord.parse('327m0002            CPU1  -AY30       A01X+020114Y+014930X0120Y    R090 S1      ', settings)
    assert r.pad_type == PadType.SMD_PAD
    assert r.net_name == 'm0002'
    assert r.is_connected
    assert r.ref_des == 'CPU1'
    assert not r.is_via
    assert r.pin == 'AY30'
    assert not r.is_middle
    assert r.hole_dia is None
    assert r.is_plated is None
    assert r.access_layer == 1
    assert math.isclose(r.x, 20114/1000) and math.isclose(r.y, 14930/1000)
    assert math.isclose(r.w, 120/1000) and r.h is None
    assert math.isclose(r.rotation, math.pi/2)
    assert r.solder_mask == SoldermaskInfo.PRIMARY
    assert r.unit == settings.unit

    r = TestRecord.parse('327m0002            U15   -D3         A01X+011545Y+003447X0090Y    R090 S1      ', settings)
    assert r.pad_type == PadType.SMD_PAD
    assert r.net_name == 'm0002'
    assert r.is_connected
    assert r.ref_des == 'U15'
    assert r.pin == 'D3'
    assert not r.is_middle
    assert r.hole_dia is None
    assert r.is_plated is None
    assert r.access_layer == 1
    assert math.isclose(r.w, 90/1000) and r.h is None

    r = TestRecord.parse('327VSUMPG           C39   -2   M      A01X+013050Y+020050X0350Y0320R270 S1      ', settings)
    assert r.pad_type == PadType.SMD_PAD
    assert r.net_name == 'VSUMPG'
    assert r.is_connected
    assert r.ref_des == 'C39'
    assert r.pin == '2'
    assert r.is_middle
    assert r.hole_dia is None
    assert r.is_plated is None
    assert r.access_layer == 1
    assert math.isclose(r.w, 350/1000) and math.isclose(r.h, 320/1000)
    assert math.isclose(r.rotation, math.pi*3/2)

    r = TestRecord.parse('327N/C              CPU1  -AD2        A01X+023191Y+020393X0110Y    R090 S1      ', settings)
    assert r.pad_type == PadType.SMD_PAD
    assert r.net_name == None
    assert not r.is_connected
    assert r.ref_des == 'CPU1'
    assert r.pin == 'AD2'
    assert r.hole_dia is None
    assert r.is_plated is None
    assert r.access_layer == 1
    assert math.isclose(r.w, 110/1000) and r.h is None

    r = TestRecord.parse('317m0002            VIA   -    MD0080PA00X+011900Y+004000X0160Y         S3      ', settings)
    assert r.pad_type == PadType.THROUGH_HOLE
    assert r.net_name == 'm0002'
    assert r.is_connected
    assert r.ref_des is None
    assert r.is_via
    assert r.pin is None
    assert r.is_middle
    assert r.hole_dia == 80/1000
    assert r.is_plated
    assert r.access_layer == 0
    assert math.isclose(r.w, 160/1000) and r.h is None
    assert r.rotation == 0
    assert r.solder_mask == SoldermaskInfo.BOTH

    r = TestRecord.parse('317GND              VIA   -    MD0080PA00X+023800Y+010100X0160Y         S0      ', settings)
    assert r.pad_type == PadType.THROUGH_HOLE
    assert r.net_name == 'GND'
    assert r.is_connected
    assert r.is_via
    assert r.pin is None
    assert r.hole_dia == 80/1000
    assert r.is_plated
    assert r.access_layer == 0
    assert r.solder_mask == SoldermaskInfo.NONE

def test_record_idempotence():
    records = [
            '327m0002            CPU1  -AY30       A01X+020114Y+014930X0120Y    R090 S1      ',
            '327m0002            U15   -D3         A01X+011545Y+003447X0090Y    R090 S1      ',
            '327VSUMPG           C39   -2   M      A01X+013050Y+020050X0350Y0320R270 S1      ',
            '317m0002            VIA   -    MD0080PA00X+011900Y+004000X0160Y         S3      ',
            '317GND              VIA   -    MD0080PA00X+023800Y+010100X0160Y         S0      ',]

    for unit in MM, Inch:
        settings = FileSettings(unit=unit)
        for record in records:
            ra = TestRecord.parse(record, settings)
            a = list(ra.format(settings))[0]
            rb = TestRecord.parse(a, settings)
            b = list(rb.format(settings))[0]
            print('ra', ra)
            print('rb', rb)
            print('0', record)
            print('a', a)
            print('b', b)
            assert a == b




#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2022 Jan Götte <code@jaseg.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pathlib import Path

import pytest

from .utils import *
from ..layers import LayerStack
from ..rs274x import GerberFile
from ..excellon import ExcellonFile

# hand-classified
REFERENCE_DIRS = {
    'Target3001': {
        'IRNASIoTbank1.2.Apr': None,
        'IRNASIoTbank1.2.Bot': 'bottom copper',
        'IRNASIoTbank1.2.Drill': 'drill plated',
        'IRNASIoTbank1.2.Info': None,
        'IRNASIoTbank1.2.Outline': 'mechanical outline',
        'IRNASIoTbank1.2.PasteBot': 'bottom paste',
        'IRNASIoTbank1.2.PasteTop': 'top paste',
        'IRNASIoTbank1.2.PosiBot': 'bottom silk',
        'IRNASIoTbank1.2.PosiTop': 'top silk',
        'IRNASIoTbank1.2.StopBot': 'bottom mask',
        'IRNASIoTbank1.2.StopTop': 'top mask',
        'IRNASIoTbank1.2.Tool': None,
        'IRNASIoTbank1.2.Top': 'top copper',
        'IRNASIoTbank1.2.Whl': None,
        },

    'allegro': {
        '08_057494d-ipc356.ipc': 'other netlist',
        '08_057494d.rou': 'mechanical outline',
        'Read_Me.1': None,
        'art_param.txt': None,
        'assy1.art': None,
        'assy2.art': None,
        'fab1.art': None,
        'l1_primary.art': 'top copper',
        'l2_gnd.art': 'inner_2 copper',
        'l3_vcc.art': 'inner_3 copper',
        'l4_secondary.art': 'bottom copper',
        'mask_prm.art': 'top mask',
        'mask_sec.art': 'bottom mask',
        'nc_param.txt': None,
        'ncdrill-1-4.drl': 'drill unknown',
        'ncdrill.log': None,
        'netlist.err': None,
        'paste_prm.art': 'top paste',
        'paste_sec.art': 'bottom paste',
        'photo.log': None,
        'silk_prm.art': 'top silk',
        'silk_sec.art': 'bottom silk',
        },

    'allegro-2': {
        'MINNOWMAX_REVA2_PUBLIC_BOTTOMSIDE.pdf': None,
        'MINNOWMAX_REVA2_PUBLIC_TOPSIDE.pdf': None,
        'MinnowMax_RevA1_IPC356A.ipc': 'other netlist',
        'MinnowMax_RevA1_DRILL/MinnowMax_RevA1_NCDRILL.drl': 'drill unknown',
        'MinnowMax_RevA1_DRILL/MinnowMax_RevA1_NCROUTE.rou': 'drill unknown',
        'MinnowMax_RevA1_DRILL/nc_param.txt': None,
        'MinnowMax_RevA1_DRILL/ncdrill.log': None,
        'MinnowMax_RevA1_DRILL/ncroute.log': None,
        'MinnowMax_assy.art': None,
        'MinnowMax_bslk.art': 'bottom silk',
        'MinnowMax_fab.art': None,
        'MinnowMax_lyr10_GAF.art': 'bottom copper',
        'MinnowMax_lyr1_GAF.art': 'top copper',
        'MinnowMax_lyr2.art': 'inner_2 copper',
        'MinnowMax_lyr3.art': 'inner_3 copper',
        'MinnowMax_lyr4.art': 'inner_4 copper',
        'MinnowMax_lyr5.art': 'inner_5 copper',
        'MinnowMax_lyr6.art': 'inner_6 copper',
        'MinnowMax_lyr7.art': 'inner_7 copper',
        'MinnowMax_lyr8.art': 'inner_8 copper',
        'MinnowMax_lyr9.art': 'inner_9 copper',
        'MinnowMax_smc_GAF.art': 'top mask',
        'MinnowMax_sms_GAF.art': 'bottom mask',
        'MinnowMax_spc.art': 'top paste',
        'MinnowMax_sps.art': 'bottom paste',
        'MinnowMax_tslk_GAF.art': 'top silk',
        },

    'altium-composite-drill': {
        'Gerber/LimeSDR-QPCIe_1v2-macro.APR_LIB': None,
        'Gerber/LimeSDR-QPCIe_1v2.EXTREP': None,
        'Gerber/LimeSDR-QPCIe_1v2.G1': 'inner_1 copper',
        'Gerber/LimeSDR-QPCIe_1v2.G10': 'inner_10 copper',
        'Gerber/LimeSDR-QPCIe_1v2.G11': 'inner_11 copper',
        'Gerber/LimeSDR-QPCIe_1v2.G12': 'inner_12 copper',
        'Gerber/LimeSDR-QPCIe_1v2.G2': 'inner_2 copper',
        'Gerber/LimeSDR-QPCIe_1v2.G3': 'inner_3 copper',
        'Gerber/LimeSDR-QPCIe_1v2.G4': 'inner_4 copper',
        'Gerber/LimeSDR-QPCIe_1v2.G5': 'inner_5 copper',
        'Gerber/LimeSDR-QPCIe_1v2.G6': 'inner_6 copper',
        'Gerber/LimeSDR-QPCIe_1v2.G7': 'inner_7 copper',
        'Gerber/LimeSDR-QPCIe_1v2.G8': 'inner_8 copper',
        'Gerber/LimeSDR-QPCIe_1v2.G9': 'inner_9 copper',
        'Gerber/LimeSDR-QPCIe_1v2.GBL': 'bottom copper',
        'Gerber/LimeSDR-QPCIe_1v2.GBO': 'bottom silk',
        'Gerber/LimeSDR-QPCIe_1v2.GBP': 'bottom paste',
        'Gerber/LimeSDR-QPCIe_1v2.GBS': 'bottom mask',
        'Gerber/LimeSDR-QPCIe_1v2.GM1': 'mechanical outline',
        'Gerber/LimeSDR-QPCIe_1v2.GM14': None,
        'Gerber/LimeSDR-QPCIe_1v2.GM15': None,
        'Gerber/LimeSDR-QPCIe_1v2.GPB': None,
        'Gerber/LimeSDR-QPCIe_1v2.GPT': None,
        'Gerber/LimeSDR-QPCIe_1v2.GTL': 'top copper',
        'Gerber/LimeSDR-QPCIe_1v2.GTO': 'top silk',
        'Gerber/LimeSDR-QPCIe_1v2.GTP': 'top paste',
        'Gerber/LimeSDR-QPCIe_1v2.GTS': 'top mask',
        'Gerber/LimeSDR-QPCIe_1v2.REP': None,
        'Gerber/LimeSDR-QPCIe_1v2.RUL': None,
        'Gerber/LimeSDR-QPCIe_1v2.apr': None,
        'NC Drill/LimeSDR-QPCIe_1v2-RoundHoles.TXT': 'drill unknown',
        'NC Drill/LimeSDR-QPCIe_1v2-SlotHoles.TXT': 'drill unknown',
        'NC Drill/LimeSDR-QPCIe_1v2.DRR': None,
        'NC Drill/LimeSDR-QPCIe_1v2.LDP': None,
        },

# TODO there are three designs in this folder. make test work with that. 
#     'diptrace': {
#         'mainboard.drl': 'drill plated',
#         'mainboard_BoardOutline.gbr': 'mechanical outline',
#         'mainboard_Bottom.gbr': 'bottom copper',
#         'mainboard_BottomMask.gbr': 'bottom mask',
#         'mainboard_Top.gbr': 'top copper',
#         'mainboard_TopMask.gbr': 'top mask',
#         'mainboard_TopSilk.gbr': 'top silk',
#         },

    'eagle-newer': {
        'copper_bottom.gbr': 'bottom copper',
        'copper_top.gbr': 'top copper',
        'drills.xln': 'drill unknown',
        'gerber_job.gbrjob': None,
        'profile.gbr': 'mechanical outline',
        'silkscreen_bottom.gbr': 'bottom silk',
        'silkscreen_top.gbr': 'top silk',
        'soldermask_bottom.gbr': 'bottom mask',
        'soldermask_top.gbr': 'top mask',
        'solderpaste_bottom.gbr': 'bottom paste',
        'solderpaste_top.gbr': 'top paste',
        },

    'eagle_files': {
        'copper_bottom_l4.gbr': 'bottom copper',
        'copper_inner_l2.gbr': 'inner_2 copper',
        'copper_inner_l3.gbr': 'inner_3 copper',
        'copper_top_l1.gbr': 'top copper',
        'profile.gbr': 'mechanical outline',
        'silkscreen_bottom.gbr': 'bottom silk',
        'silkscreen_top.gbr': 'top silk',
        'soldermask_bottom.gbr': 'bottom mask',
        'soldermask_top.gbr': 'top mask',
        'solderpaste_bottom.gbr': 'bottom paste',
        'solderpaste_top.gbr': 'top paste',
        },

    'easyeda': {
        'Gerber_BoardOutline.GKO': 'mechanical outline',
        'Gerber_BottomLayer.GBL': 'bottom copper',
        'Gerber_BottomSolderMaskLayer.GBS': 'bottom mask',
        'Gerber_Drill_NPTH.DRL': 'drill nonplated',
        'Gerber_Drill_PTH.DRL': 'drill plated',
        'Gerber_TopLayer.GTL': 'top copper',
        'Gerber_TopPasteMaskLayer.GTP': 'top paste',
        'Gerber_TopPasteMaskLayer.bottom.svg': None,
        'Gerber_TopPasteMaskLayer.gtp.top.solderpaste.svg': None,
        'Gerber_TopPasteMaskLayer.gtp.top.solderpaste_2.svg': None,
        'Gerber_TopPasteMaskLayer.top.svg': None,
        'Gerber_TopSilkLayer.GTO': 'top silk',
        'Gerber_TopSolderMaskLayer.GTS': 'top mask',
        'How-to-order-PCB.txt': None,
        },

    'fritzing': {
        'combined.GKO': 'mechanical outline',
        'combined.gbl': 'bottom copper',
        'combined.gbo': 'bottom silk',
        'combined.gbs': 'bottom mask',
        'combined.gm1': None,
        'combined.gtl': 'top copper',
        'combined.gto': 'top silk',
        'combined.gts': 'top mask',
        'combined.txt': 'drill unknown',
        'gyro_328p_6050_2021_panelize.gerberset': None,
        },

# same as above, two designs in one folder
#    'geda': {
#        'controller.bottom.gbr': 'bottom copper',
#        'controller.bottommask.gbr': 'bottom mask',
#        'controller.fab.gbr': None,
#        'controller.group3.gbr': None,
#        'controller.plated-drill.cnc': 'drill plated',
#        'controller.top.gbr': 'top copper',
#        'controller.topmask.gbr': 'top mask',
#        'controller.topsilk.gbr': 'top silk',
#        'controller.unplated-drill.cnc': 'drill nonplated',
#        },

    'pcb-rnd': {
        'power-art.asb': None,
        'power-art.ast': None,
        'power-art.fab': None,
        'power-art.gbl': 'bottom copper',
        'power-art.gbo': 'bottom silk',
        'power-art.gbp': 'bottom paste',
        'power-art.gbs': 'bottom mask',
        'power-art.gko': 'mechanical outline',
        'power-art.gtl': 'top copper',
        'power-art.gto': 'top silk',
        'power-art.gtp': 'top paste',
        'power-art.gts': 'top mask',
        'power-art.lht': None,
        'power-art.xln': 'drill unknown',
        },

    'siemens': {
        '80101_0125_F200_ContourPlated.ncd': 'mechanical outline',
        '80101_0125_F200_DrillDrawingThrough.gdo': None,
        '80101_0125_F200_L01_Top.gdo': 'top copper',
        '80101_0125_F200_L02.gdo': 'inner_2 copper',
        '80101_0125_F200_L03.gdo': 'inner_3 copper',
        '80101_0125_F200_L04.gdo': 'inner_4 copper',
        '80101_0125_F200_L05.gdo': 'inner_5 copper',
        '80101_0125_F200_L06.gdo': 'inner_6 copper',
        '80101_0125_F200_L07.gdo': 'inner_7 copper',
        '80101_0125_F200_L08.gdo': 'inner_8 copper',
        '80101_0125_F200_L09.gdo': 'inner_9 copper',
        '80101_0125_F200_L10.gdo': 'inner_10 copper',
        '80101_0125_F200_L11.gdo': 'inner_11 copper',
        '80101_0125_F200_L12_Bottom.gdo': 'bottom copper',
        '80101_0125_F200_SilkscreenBottom.gdo': 'bottom silk',
        '80101_0125_F200_SilkscreenTop.gdo': 'top silk',
        '80101_0125_F200_SolderPasteBottom.gdo': 'bottom paste',
        '80101_0125_F200_SolderPasteTop.gdo': 'top paste',
        '80101_0125_F200_SoldermaskBottom.gdo': 'bottom mask',
        '80101_0125_F200_SoldermaskTop.gdo': 'top mask',
        '80101_0125_F200_ThruHoleNonPlated.ncd': 'drill nonplated',
        '80101_0125_F200_ThruHolePlated.ncd': 'drill plated',
        },

    'siemens-2': {
        'Gerber/BoardOutlline.gdo': 'mechanical outline',
        'Gerber/DrillDrawingThrough.gdo': None,
        'Gerber/EtchLayerBottom.gdo': 'bottom copper',
        'Gerber/EtchLayerTop.gdo': 'top copper',
        'Gerber/GerberPlot.gpf': None,
        'Gerber/PCB.dsn': None,
        'Gerber/SolderPasteBottom.gdo': 'bottom paste',
        'Gerber/SolderPasteTop.gdo': 'top paste',
        'Gerber/SoldermaskBottom.gdo': 'bottom mask',
        'Gerber/SoldermaskTop.gdo': 'top mask',
        'NCDrill/ContourPlated.ncd': 'mechanical outline',
        'NCDrill/ThruHoleNonPlated.ncd': 'drill nonplated',
        'NCDrill/ThruHolePlated.ncd': 'drill plated',
        },

    'upverter': {
        'design_export.drl': 'drill unknown',
        'design_export.gbl': 'bottom copper',
        'design_export.gbo': 'bottom silk',
        'design_export.gbp': 'bottom paste',
        'design_export.gbs': 'bottom mask',
        'design_export.gko': 'mechanical outline',
        'design_export.gtl': 'top copper',
        'design_export.gto': 'top silk',
        'design_export.gtp': 'top paste',
        'design_export.gts': 'top mask',
        'design_export.xln': 'drill unknown',
        'layers.cfg': None,
        },

    'zuken-emulated': {
        'Gerber/MetalMask-A.fph': 'top paste',
        'Gerber/MetalMask-B.fph': 'bottom paste',
        'Gerber/Symbol-A.fph': 'top silk',
        'Gerber/Symbol-B.fph': 'bottom silk',
        'Gerber/Resist-A.fph': 'top mask',
        'Gerber/Resist-B.fph': 'bottom mask',
        'Gerber/Conductive-1.fph': 'top copper',
        'Gerber/Conductive-2.fph': 'bottom copper',
        'Drill/8seg_Driver__routed_Drill_thru_plt.fdr/8seg_Driver__routed_Drill_thru_plt.fdl': None,
        'Drill/8seg_Driver__routed_Drill_thru_plt.fdr/8seg_Driver__routed_Drill_thru_plt.fdr': 'drill plated',
        'Drill/8seg_Driver__routed_Drill_thru_nplt.fdr': 'drill nonplated',
        }
    }

@filter_syntax_warnings
@pytest.mark.parametrize('ref_dir', list(REFERENCE_DIRS.items()))
def test_layer_classifier(ref_dir):
    ref_dir, file_map = ref_dir
    path = reference_path(ref_dir)
    print('Reference path is', path)
    file_map = { filename: role for filename, role in file_map.items() if role is not None } 
    rev_file_map = { tuple(value.split()): key for key, value in file_map.items() }
    drill_files = { filename: role for filename, role in file_map.items() if role.startswith('drill') }

    stack = LayerStack.from_directory(path)
    print('loaded layers:', ', '.join(f'{side} {use}' for side, use in stack.graphic_layers))

    for side in 'top', 'bottom':
        for layer in 'copper', 'silk', 'mask', 'paste':
            if 'allegro-2' in ref_dir and layer in ('silk', 'mask', 'paste'):
                # This particular example has very poorly named files
                continue
            if 'easyeda' in ref_dir and layer == 'paste' and side == 'bottom':
                continue

            if (side, layer) in rev_file_map:
                assert (side, layer) in stack
                found = stack[side, layer]
                assert isinstance(found, GerberFile)
                assert found.original_path.name == Path(rev_file_map[side, layer]).name

            else: # not in file_map
                assert (side, layer) not in stack

    assert len(stack.drill_layers) == len(drill_files)

    for filename, role in drill_files.items():
        print('drill:', filename, role)
        print([(layer.original_path, layer.original_path == Path(filename).name) for layer in stack.drill_layers]) 
        assert any(layer.original_path.name == Path(filename).name for layer in stack.drill_layers) 

    for layer in stack.drill_layers:
        if 'upverter' not in ref_dir:
            assert isinstance(layer, ExcellonFile)

